(******************************************************************************
 * Copyright (c) 2012, Toyama&Aoto Laboratory, Tohoku University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(******************************************************************************
 * file: rwtools/rwchecker/cr_solver.sml
 * description: confluence checker
 * author: YOSHIDA Junichi
 * author: AOTO Takahito
 * 
 * $Id: cr_solver.sml,v 1.28 2012/05/14 04:48:14 aoto Exp $
 ******************************************************************************)

signature CR_SOLVER = 
sig
    val runProfile: bool ref
    val runDebug: bool ref

    type options = {
         useDirect:bool,   (*  $BD>OBJ,2r$r;H$&$+(B *)
         useCommutative:bool,   (*  $B2D49J,2r$r;H$&$+(B *)
         useNonMinimalCommutative:bool,   (* $BHs6K>.2D49J,2r$r;H$&$+(B *)
         useLayer:bool    (*  $BAXJ]B8J,2r$r;H$&$+(B *)
    }

    val crSolver: Solver.options
                  -> DpSolver.options
                  -> PoSolver.options
                  -> options
                  -> (Term.term * Term.term) list 
                  ->  Cr.ConfluenceResult

    val useHuetToyamaOostrom: bool ref;
    val useHuetStrongClosed: bool ref;
    val useOyamaguchiOhta: bool ref;
    val useToyamaParallel: bool ref;
    val useOkuiSimultaneous: bool ref;
    val useDecreasingDiagrams: bool ref;
    val useOhtaOyamaguchiToyama: bool ref;
    val useGomiOyamaguchiOhta: bool ref;
    val useNonConfluence: bool ref;
    val useKnuthBendix: bool ref;
    val useHuetModulo: bool ref;
    val useBidirectionalCompletion: bool ref;

end

structure CrSolver : CR_SOLVER =
struct
    local
	open Prop
	open Cr
	structure A = Atom
	structure L = List
	structure LU = ListUtil
	structure FS = FunSet
	structure ILM = IntListMap2
	structure TS = TermSet
	structure VS = VarSet
	open PrintUtil
    in
    
    exception CrSolverError
          
    val runProfile = ref false : bool ref
    val runDebug = ref false : bool ref
    fun debug f = if !runDebug then f () else ()

    type options = {
         useDirect:bool,   (*  $BD>OBJ,2r$r;H$&$+(B *)
         useCommutative:bool,   (*  $B2D49J,2r$r;H$&$+(B *)
         useNonMinimalCommutative:bool,   (* $BHs6K>.2D49J,2r$r;H$&$+(B *)
         useLayer:bool    (*  $BAXJ]B8J,2r$r;H$&$+(B *)
    }

    val useHuetToyamaOostrom = ref true
    val useHuetStrongClosed = ref true
    val useOyamaguchiOhta = ref true
    val useToyamaParallel = ref true
    val useOkuiSimultaneous = ref true
    val useDecreasingDiagrams = ref true
    val useOhtaOyamaguchiToyama = ref true
    val useGomiOyamaguchiOhta = ref true
    val useNonConfluence = ref true
    val useKnuthBendix = ref true
    val useHuetModulo = ref false (* included in BidirectionalCompletion *)
    val useBidirectionalCompletion = ref true

    val nameHuetToyamaOostrom = "Development Closed"
    val nameHuetStrongClosed = "Strongly Closed"
    val nameOyamaguchiOhta = "Upside-Parallel-Closed/Outside-Closed"
    val nameToyamaParallel = "Toyama (Parallel CPs)"
    val nameOkuiSimultaneous = "Okui (Simultaneous CPs)"
    val nameHuetModulo = "Huet (modulo AC)"
    val nameDecreasingDiagrams = "Diagram Decreasing"
    val nameOhtaOyamaguchiToyama = "Simple-Right-Linear"
    val nameGomiOyamaguchiOhta = "Strongly Depth-Preserving & Non-E-Overlapping"
    val nameNonConfluence = "Non-Confluence"
    val nameKnuthBendix = "Knuth&Bendix"
    fun unknown s = "unknown " ^ s

    local 
	open Term
	open Cr
	open PrintUtil
  (* $B9gN.@-$N4pK\H=Dj(B *)
    in
   fun checkConfluenceConditions2 smtSolver tmpDir 
				  isTerminating isRelativeTerminating rs = 
       let 
	   val _ = print "Apply Direct Methods...\n"


	   val inCpeaksWithIndexAndPosFs = Cr.insideCriticalPeaksWithIndexAndPosFs rs

	   val inCpsPos = L.map (fn (mn,(ps,fs,x,y,z)) => (ps,y,z)) inCpeaksWithIndexAndPosFs
	   val inCpeaks = L.map (fn (mn,(ps,fs,x,y,z)) => (x,y,z)) inCpeaksWithIndexAndPosFs
	   val outCpeaks = outsideCriticalPeaksInOneside rs

	   val inCps = L.map (fn (_,y,z) => (y,z)) inCpeaks
	   val outCps = L.map (fn (_,y,z) => (y,z)) outCpeaks

	   val _ = print ("Inner CPs:\n" ^ (LU.toStringCommaLnSquare (IOFotrs.prEq) inCps))
	   val _ = print ("Outer CPs:\n" ^ (LU.toStringCommaLnSquare (IOFotrs.prEq) outCps))

(*
	   val inCpeaksWithIndexAndPosFs = []
	   val inCpsPos = []
	   val inCpeaks = []
	   val outCpeaks = []
	   val inCps = []
	   val outCps = []
*)

	   val (_,dSymSet,cSymSet,_) = Trs.fdcSetInRules rs

	   val maxRewriteLenForNonConfluenceCandidates = ref 2

	   val developSetTable = ref TermMap.empty

	   fun insertDevelopSet term = 
	       let val tset = Rewrite.developOneStepReductSet rs term
		   val _ = developSetTable := TermMap.insert (!developSetTable,term,tset)
	       in tset
	       end

	   fun lookupDevelopSet term =
	       case TermMap.find (!developSetTable,term) of
		   SOME tset => tset
		 | NONE => insertDevelopSet term

	   val reductSetTable = ref TermMap.empty

	   fun insertReductSet term = 
	       let val maxLen = !maxRewriteLenForNonConfluenceCandidates
		   val tset = Rewrite.manyStepsReductSet rs maxLen term
		   val _ = reductSetTable := TermMap.insert (!reductSetTable,term,tset)
	       in tset
	       end

	   fun lookupReductSet term =
	       case TermMap.find (!reductSetTable,term) of
		   SOME tset => tset
		 | NONE => insertReductSet term

	   fun checkHuetToyamaOostrom _ = 
	       (!useHuetToyamaOostrom)
	       andalso
	       let fun checkIn (l,r) = Term.equal (l,r) 
				       orelse let val dl = lookupDevelopSet l
						  val _ = debug (fn _ => println ("check inner CP " ^ (Trs.prEq (l,r))))
						  val _ = debug (fn _ => println ("{ s | l -o-> s } = " ^
								 (prSetInOneLine Term.toString (TS.listItems dl))))
					      in TS.member (dl,r)
					      end
						  
		   fun checkOut (l,r) = 
		       Term.equal (l,r) 
		       orelse 
		       let val l1 = lookupDevelopSet l
			   val r1 = lookupDevelopSet r
			   val _ = debug (fn _ => println ("check outer CP " ^ (Trs.prEq (l,r))))
			   val _ = debug (fn _ => println ("X1 = { s | l -o-> s } = " ^ 
						  (prSetInOneLine Term.toString (TS.listItems l1))))
			   val _ = debug (fn _ => println ("Y1 = { t | r -o-> t } = " ^
						  (prSetInOneLine Term.toString (TS.listItems r1))))
			   val _ = debug (fn _ => println ("check X1 cap Y1 /= {} "))
		       in not (TS.isEmpty (TS.intersection (l1,r1)))
			  orelse let val l2 = lookupReductSet l
				     val r2 = lookupReductSet r
				     val _ = debug (fn _ => println ("X2 = { s | l -n-> s } = " ^
						    (prSetInOneLine Term.toString (TS.listItems l2))))
				     val _ = debug (fn _ => println ("Y2 = { s | r -n-> s } = " ^
								   (prSetInOneLine Term.toString (TS.listItems r2))))
				     val _ = debug (fn _ => println ("check X1 cap Y2 /= {} and X2 cap Y1 /= {}"))
				 in not (TS.isEmpty (TS.intersection (l1,r2)))
				    andalso not (TS.isEmpty (TS.intersection (l2,r1)))
				 end
		       end
	       in
		   if L.all checkIn inCps
		      andalso L.all checkOut outCps
		   then (println nameHuetToyamaOostrom; true)
		   else (println (unknown nameHuetToyamaOostrom); false)
	       end

	   fun checkHuetStrongClosed _ = 
	       (!useHuetStrongClosed)
	       andalso
	       let fun strongClosed (l,r) = 
		       Term.equal (l,r) 
		       orelse 		    
		       let 
			   val _ = debug (fn _ => println ("check CP " ^ (Trs.prEq (l,r))))
			   val r1 = lookupDevelopSet r
			   val l1 = TS.add (Rewrite.oneStepReductSet rs l, l)
			   val _ = debug (fn _ => println ("X = { s | l -=-> s } = " ^
							   (prSetInOneLine Term.toString (TS.listItems l1))))
			   val _ = debug (fn _ => println ("Y1 = { s | r -o-> s } = " ^
							   (prSetInOneLine Term.toString (TS.listItems r1))))
			   val _ = debug (fn _ => println ("check X cap Y1 /= {} "))
		       in not (TS.isEmpty (TS.intersection (r1,l1)))
			  orelse
			  let val r2 = lookupReductSet r
			   val _ = debug (fn _ => println ("Y2 = { s | r -n-> s } = " ^
							   (prSetInOneLine Term.toString (TS.listItems r2))))
			   val _ = debug (fn _ => println ("check X cap Y2 /= {} "))
			  in not (TS.isEmpty (TS.intersection (r2,l1)))
			  end
		       end
	       in
		   if L.all strongClosed outCps
		      andalso L.all strongClosed (L.map (fn (x,y) => (y,x)) outCps)
		      andalso L.all strongClosed inCps
		      andalso L.all strongClosed (L.map (fn (x,y) => (y,x)) inCps)
		   then (println nameHuetStrongClosed; true)
		   else (println (unknown nameHuetStrongClosed);false)
	       end


	   fun checkOyamaguchiOhta _ = 
	       (!useOyamaguchiOhta)
	       andalso
	       let val depthRestrectedClosedCps = isDepthRestrictedClosedCps rs inCpsPos outCps
		   val _ = if depthRestrectedClosedCps
			   then println nameOyamaguchiOhta
			   else println ("unknown " ^ nameOyamaguchiOhta)
	       in depthRestrectedClosedCps
	       end


	   fun checkToyamaParallel _ = 
	       (!useToyamaParallel)
	       andalso
	       let val parallelCriticalPairClosed = 
		       isParallelCriticalPairClosed rs (inCps @ outCps) 
		   val _ = if parallelCriticalPairClosed
			   then println nameToyamaParallel
			   else println (unknown nameToyamaParallel)
	       in parallelCriticalPairClosed
	       end

	   fun checkOkuiSimultaneous _ = 
	       (!useOkuiSimultaneous)
	       andalso
	       let val okuiClosedCps = isOkuiClosedCps rs
		   val _ = if okuiClosedCps
			   then println nameOkuiSimultaneous
			   else println (unknown nameOkuiSimultaneous)
	       in okuiClosedCps
	       end

	   fun checkHuetModulo _ = 
	       (!useHuetModulo)
	       andalso
	       if Cr.isHuetModulo isRelativeTerminating rs
	       then (println nameHuetModulo; true)
	       else (println (unknown nameHuetModulo);false)
	       
	   fun checkDecreasingDiagrams _ = 
	       (!useDecreasingDiagrams)
	       andalso
	       let val _ = print "check Locally Decreasing Diagrams by Rule Labelling...\n"
(*		   val decreasingCps = CrDiagram2.hasDecreasingCps0 smtSolver tmpDir rs *)
(*		   val decreasingCps = CrDiagram2.hasDecreasingCps smtSolver tmpDir rs  *)
		   val decreasingCps = CrDiagram.hasDecreasingCps smtSolver tmpDir rs  
		   val _ = if decreasingCps
			   then println nameDecreasingDiagrams
			   else println (unknown nameDecreasingDiagrams)
	       in decreasingCps
	       end

	   fun checkOhtaOyamaguchiToyama _ =
	       (!useOhtaOyamaguchiToyama)
	       andalso
	       let fun checkVar (l,r) = 
		       VS.isEmpty
			   (VS.intersection (		      
			    VS.difference (Term.varSetInTerm l, Term.linearVarSetInTerm l),
			    Term.varSetInTerm r))
		   val _ = debug (fn _ => println "check all x in V(r) is linear in l")
	       in if L.all checkVar rs
		     andalso 
		     let val cp = Cr.criticalPairs (L.map (fn (l,r) => (Subst.linearize l,r)) rs)
			 val _ = debug (fn _ => print ("CPs after linearization:\n"))
			 val _ = debug (fn _ => print (LU.toStringCommaLnSquare (IOFotrs.prEq) cp))
		     in  null cp
		     end
		  then (println nameOhtaOyamaguchiToyama; true)
		  else (println (unknown nameOhtaOyamaguchiToyama); false)
	       end


	   fun checkGomiOyamaguchiOhta _ = 
	       (!useGomiOyamaguchiOhta)
	       andalso
	       let fun varPosInTerm vset pos (Var (x,_)) = 
		       if VS.member (vset, x) 
		       then [(length pos,x)]
		       else []
		     | varPosInTerm vset pos (Fun (_,ts,_)) = 
		       LU.mapAppend 
			   (fn i => varPosInTerm vset (pos @ [i+1]) (L.nth (ts,i)))
			   (List.tabulate (length ts, fn x=>x))
		   fun checkVar (l,r) = 
		       let val vset = VS.intersection (varSetInTerm l, varSetInTerm r)
			   val lv = varPosInTerm vset [] l
			   val rv = varPosInTerm vset [] r
			   fun refineLv [] zs = zs
			     | refineLv ((i,x)::ys) zs =
			       refineLv (L.filter (fn (j,y) => not (Var.equal (x,y)) 
							       orelse j <= i) ys)
					((i,x)::(L.filter (fn (j,y) => not (Var.equal (x,y)) 
								       orelse j <= i) zs))
			   fun checkVar [] _ = true
			     | checkVar ((i,x)::ys) zs = 
			       L.all (fn (j,z) => not (Var.equal (x,z))
						  orelse i >= j) zs
			       andalso checkVar ys zs
		       in checkVar (refineLv lv []) rv
		       end
		   val _ = debug (fn _ => println "check all x in V(r), max {|p| | r/p = x } <= min {|q| | l/q = x}")
	       in if L.all checkVar rs
		     andalso 
		     let val cp = Cr.criticalPairs (L.map (fn (l,r) => (Subst.linearize l,r)) rs)
			 val _ = debug (fn _ => print ("check non-E-overlapping (approximated):\n"))
			 val _ = debug (fn _ => print ("CPs after linearization:\n"))
			 val _ = debug (fn _ => print (LU.toStringCommaLnSquare (IOFotrs.prEq) cp))
		     in  null cp
		     end
		  then  (println nameGomiOyamaguchiOhta; true)
		  else (println (unknown nameGomiOyamaguchiOhta); false)
	       end

(* 	   local exception UseAll *)
(* 	   in *)
(* 	   fun takeFunSetClosure fSet1 =  *)
(* 	       let  *)
(* 		   val fSet2 = L.foldr (fn ((l,r),fset) =>  *)
(* 					   case Term.funRootOfTerm l  *)
(* 					    of SOME f => if FS.member (fset,f)  *)
(* 							 then case Term.funRootOfTerm r *)
(* 								       of NONE => raise UseAll *)
(* 									| SOME g => FS.add (fset,g) *)
(* 							 else fset *)
(* 					     | NONE => fset (\* should not be here *\)) *)
(* 				       fSet1 *)
(* 				       rs *)
(* 	       in if FS.isSubset (fSet2,fSet1) *)
(* 		  then SOME fSet1 *)
(* 		  else takeFunSetClosure fSet2 *)
(* 	       end *)
(* 	       handle UseAll => NONE *)
(* 	   end *)

(* 	   (\*** ineffective ***\) *)
(* 	   fun checkByUsableHeads (s,t) =  *)
(* 	       let val sHeads = case Term.funRootOfTerm s *)
(* 				 of NONE => NONE *)
(* 				  | SOME f => takeFunSetClosure (FS.singleton f) *)
(* 		   val tHeads = case Term.funRootOfTerm t *)
(* 				 of NONE => NONE *)
(* 				  | SOME f => takeFunSetClosure (FS.singleton f) *)
(* 		   val _ = print "\n"; *)
(* 		   val _ = print (ListUtil.toStringCommaAngle Term.toString [s,t]) *)
(* 		   fun debugOut X = print (ListUtil.toStringCommaCurly *)
(* 					       Fun.toString  *)
(* 					       (FS.listItems X)) *)
(* 	       in case (sHeads, tHeads) of *)
(* 		      (SOME X, SOME Y) => let val _ = debugOut X *)
(* 					      val _ = debugOut Y *)
(* 					  in *)
(* 					      FS.isEmpty (FS.intersection (X,Y)) *)
(* 					  end *)
(* 		    | (_, _) => (print "NONE"; false) *)
(* 	       end *)



	   val depthOfMinimalForm = ref 5

	   fun directNonConfluenceCheck (s,t) = 
	       not (Term.equal (s,t))
	       andalso
	       ((Rewrite.isHeadStable rs s
		 andalso Rewrite.isHeadStable rs t
		 andalso not (Term.haveSameRoots (s,t)))
		orelse
		(case (Rewrite.minimalReductSet rs (!depthOfMinimalForm) s,  
		       Rewrite.minimalReductSet rs (!depthOfMinimalForm) t) of
		     (SOME X, SOME Y) => TS.isEmpty (TS.intersection (X,Y))
		   | (_,_) => false)
		   (*** orelse checkByUsableHeads (s,t) ***)
	       )


	   fun recursiveNonConfluenceCheck [] = false
	     | recursiveNonConfluenceCheck ((C,u,s,t)::cps) = 
	       if directNonConfluenceCheck (s,t)
	       then (print ("Find Non-Joinable CP reducts: <"
			    ^ (Term.toString (C s)) ^ ", " 
			    ^ (Term.toString (C t)) ^ "> from "
			    ^ (Term.toString u) ^ "\n");
		     true)
	       else case (s,t) of
			(Fun (f,ts,ty), Fun (g,ss,_)) =>
			if (Fun.equal (f,g))
			   andalso (FS.member (cSymSet,f))
			 then recursiveNonConfluenceCheck
				  ((L.map (fn i => (fn x => Fun (f,LU.replaceNth (ts,i,x),ty),
						    u,
						    L.nth (ts,i),
						    L.nth (ss,i)))
					  (L.tabulate (length ts, fn x=>x)))
				   @ cps)
			 else recursiveNonConfluenceCheck cps
		       | (_,_) => recursiveNonConfluenceCheck cps


	   fun makeNonConfluentCandidatesFromCriticalPair (u,l,r) = 
	       let val ll = TS.listItems (lookupReductSet l)
		   val rr = TS.listItems (lookupReductSet r)
	       in ListXProd.mapX (fn (x,y) => (fn z=>z, u,x,y)) (ll,rr)
	       end

	   fun checkNonConfluence _ = 
	       (!useNonConfluence)
	       andalso
	       let val _ = print "check Non-Confluence..."
		   val enhancedRules = Cr.insideCriticalPairs2 
					   (L.map (fn (l,r) => (r,l)) rs, 
					    L.filter (not o Trs.isLeftLinearRule) rs)
		   val ecps = Cr.criticalPeaks2 (rs,enhancedRules)
		   (* val _ = print "\n New CPs:"
 		    val _ = L.app (fn (_,l,r) => print (Trs.prRule (l,r))) ecps
		    val _ = print "\n" *)
		   val cps = LU.eliminateDuplication' 
				 (fn ((_,l,r),(_,l',r')) => 
				     Subst.identicalModuloRenamingRule (l,r) (l',r'))
				 (outCpeaks @ inCpeaks @ ecps)
	       in if L.exists
			 (recursiveNonConfluenceCheck 
			  o makeNonConfluentCandidatesFromCriticalPair)
			 cps
		  then (print "Not Confluent\n"; true)
		  else (print "Unknown\n"; false)
	       end

	   fun checkBidirectionalCompletion _ = 
	       (!useBidirectionalCompletion)
	       andalso
	       let val _ = print "check by Reduction-Preserving Completion...\n"
		   val result = CrCompletion.bidirectionalCompletion 
				    (isTerminating false)
				    isRelativeTerminating
				    dSymSet rs
		   val _ = if result
			   then print "Reduction-Preserving Completion\n"
			   else print "unknown Reduction-Preserving Completion\n"
	       in result
	       end

	   fun checkConfluenceConditionsForLinearSystem _ = 
	       if checkHuetToyamaOostrom ()
		  orelse
		  checkHuetStrongClosed ()
		  orelse
		  checkOyamaguchiOhta ()
		  orelse
		  checkToyamaParallel ()
		  orelse
		  checkOkuiSimultaneous ()
		  orelse
		  checkGomiOyamaguchiOhta ()
		  orelse
		  checkDecreasingDiagrams ()
		  orelse
		  checkHuetModulo ()
		  orelse
		  checkBidirectionalCompletion ()
	       then (print "Direct Methods"; report CR)
	       else if checkNonConfluence ()
	       then (print "Direct Methods"; report NotCR)
	       else (print "Direct Methods"; report Unknown)

	   fun checkConfluenceConditionsForLeftLinearSystem _ = 
	       if checkHuetToyamaOostrom ()
		  orelse
		  checkOyamaguchiOhta ()
		  orelse
		  checkToyamaParallel ()
		  orelse
		  checkOkuiSimultaneous ()
		  orelse
		  checkGomiOyamaguchiOhta ()
		  orelse
		  checkDecreasingDiagrams ()
		  orelse
		  checkHuetModulo ()
		  orelse
		  checkBidirectionalCompletion ()
	       then (print "Direct Methods"; report CR)
	       else if checkNonConfluence ()
	       then (print "Direct Methods"; report NotCR)
	       else (print "Direct Methods"; report Unknown)

	   fun checkConfluenceConditionsForRightLinearSystem _ = 
	       if checkOhtaOyamaguchiToyama ()
		  orelse
		  checkGomiOyamaguchiOhta ()
		  orelse
		  checkBidirectionalCompletion ()
	       then (print "Direct Methods"; report CR)
	       else if checkNonConfluence ()
	       then (print "Direct Methods"; report NotCR)
	       else (print "Direct Methods"; report Unknown)

	   fun checkConfluenceConditionsForNonLeftRightLinearSystem _ = 
	       if checkGomiOyamaguchiOhta ()
		  orelse
		  checkBidirectionalCompletion ()
	       then (print "Direct Methods"; report CR)
	       else if checkNonConfluence ()
	       then (print "Direct Methods"; report NotCR)
	       else (print "Direct Methods"; report Unknown)

	   fun checkConfluenceConditionsForNonTerminatingSystem _ = 
	       case (Trs.areLeftLinearRules rs,
		     Trs.areRightLinearRules rs) of
		   (true,true) => (print "Linear\n";
				   checkConfluenceConditionsForLinearSystem ())
		 | (true,false) => (print "Left-Linear, not Right-Linear\n";
				    checkConfluenceConditionsForLeftLinearSystem ())
		 | (false,true) => (print "not Left-Linear, Right-Linear\n";
				    checkConfluenceConditionsForRightLinearSystem ())
		 | (false,false) => (print "not Left-Linear, not Right-Linear\n";
				     checkConfluenceConditionsForNonLeftRightLinearSystem ())

       in
	   if (!useKnuthBendix)
	   then
	       if null inCps  (* overlay *)
	       then (print "Overlay, check Innermost Termination...\n";
		     if isTerminating true rs
		    then let val _ = print "Innermost Terminating (hence Terminating)"
			     val joinableCps = isInnerJoinableCps rs outCps
			     val _ = if joinableCps
				     then println ", WCR"
				     else println ", not WCR"
			     val _ = println nameKnuthBendix
			 in
			     if joinableCps then 
				 (print "Direct Methods"; report CR)
			     else 
				 (print "Direct Methods"; report NotCR)
			 end
		    else 
			(print "unknown Innermost Terminating\n";
			 println (unknown nameKnuthBendix);
			 checkConfluenceConditionsForNonTerminatingSystem rs))
	       else (print "not Overlay, check Termination...\n";
		     if isTerminating false rs
		     then let val _ = print "Terminating"
			      val joinableCps = isJoinableCps rs (inCps @ outCps)
			      val _ = if joinableCps
				      then println ", WCR"
				      else println ", not WCR"
			      val _ = println nameKnuthBendix
			  in
			     if joinableCps then 
				 (print "Direct Methods"; report CR)
			     else 
				 (print "Direct Methods"; report NotCR)
			  end
		     else
			 (println  "unknown/not Terminating";
			  println (unknown nameKnuthBendix);
			  checkConfluenceConditionsForNonTerminatingSystem rs))
	   else
	       checkConfluenceConditionsForNonTerminatingSystem rs
       end
    end


   fun applyDirectCheck directCheck rs (n,fs) = 
       if null rs
       then report CR
       else let val _ = print "Rewrite Rules:\n"
		val _ = print (Trs.prRules rs)
		val checkCC = directCheck rs
	    in case checkCC of
		   CR => CR
		 | NotCR => NotCR
		 | Unknown => applyDecomposition directCheck rs (n,fs)
	    end
   and applyDecomposition directCheck rs (n,fs) = 
       if (null fs)
       then 
	   Unknown
       else if not (null fs)
	  andalso n = length fs
       then (print "No further decomposition possible\n\n";
	     Unknown)
       else 
	   let 
	       val fs' = (tl fs) @ [hd fs]
	       fun direct rs = applyDirectCheck directCheck rs (1,fs')
	       fun decomp rs = applyDecomposition directCheck rs (n+1,fs')
	   in case (hd fs) of
(*		0 => CrDirect.tryDirectSumDecomposition rs (direct,decomp) *)
		0 => CrDirect.tryPersistentDecomposition rs (direct,decomp)
	      | 1 => CrLayer.tryLayerPreservingDecomposition rs (direct,decomp)
	      | 2 => CrCom.tryCommutativeDecomposition rs (direct,decomp)
	      | _ => (print "Error at applyDecomposition\n";
		      raise CrSolverError)
	   end

   exception Failure



  (* $B9gN.@-%=%k%P(B *)
  fun crSolver (opt0:Solver.options) (opt1:DpSolver.options) 
	       (opt2:PoSolver.options) (opt3:options) rs =
      let
	   val trs = Trs.rulesToTrs rs
	   val _ = debug (fn _ => print "Rewrite Rules:\n") 
	   val _ = debug (fn _ => print (Trs.prRules (#Rules trs))) 
 
           val satSolver = (#satSolver opt0)
           val smtSolver = (#smtSolver opt0)
           val snProver = (#terminationProver opt0)
           val relSnProver = (#relativeTerminationProver opt0)
           val tmpDir = (#tmpDir opt0)

	   val _ = if  (#VarCond trs) then ()
		   else (print "Variable condition is not satisfied\n";
			 raise Failure)

          fun isTerminating isOverlay rs = 
	      if snProver = ""
	      then 
		  if (#useDp opt1)
		  then if isOverlay 
		       then DpSolver.dpSolverInner satSolver tmpDir opt1 opt2 rs
		       else DpSolver.dpSolver satSolver tmpDir opt1 opt2 rs
		  else PoSolver.poSolver satSolver tmpDir opt2 (rs,[])
	      else 
		  if isOverlay 
		  then Solver.sinSolver snProver tmpDir rs
		  else Solver.snSolver snProver tmpDir rs

	  fun isRelativeTerminating (rs,es) =
	      if relSnProver = ""
	      then 
		  MatrixInterpretation.relsnSolver 
		      (opt0:Solver.options) 
		      (opt1:DpSolver.options) 
		      (opt2:PoSolver.options) (rs,es)
	      else
		  (* print "relative termination proof\n";
		   print "rs:\n";
		   print (Trs.prRules rs);
		   print "es:\n";
		   print (Trs.prRules es);
		   print relSnProver;
		   print "\n"; *)
		  Solver.relsnSolver relSnProver tmpDir (rs,es)

	  fun indexOf (l,r) = 
	      let fun indexOfSub [] _ = (print "indexOf failed\n";
					 raise CrSolverError)
		    | indexOfSub ((l',r')::rest) i
		      = if Term.equal (l,l')
			   andalso Term.equal (r,r')
			then i
			else indexOfSub rest (i+1)
	      in indexOfSub rs 0 
	      end
	   fun getRules ns = L.map (fn i => L.nth (rs,i)) ns
	   val rem = ref ILM.empty

           (* $B<+A3?t%j%9%H$GF~NO(BTRS$B$NItJ,=89g$rBP1~$5$;$k(B *)
	   fun lookup xs  = 
	       case ILM.find (!rem,xs) of
		   SOME ans=> (print "(cached)"; ans)
		 | NONE => let val ans = checkConfluenceConditions2 
					     smtSolver tmpDir
					     isTerminating 
					     isRelativeTerminating 
					     (getRules xs) 
			       val _ = ILM.insert (!rem, xs, ans)
			   in ans
			   end
	   fun directCheck rs = lookup (L.map indexOf rs)

	   val _ = if #useNonMinimalCommutative opt3
		   then CrCom.useNonMinimal := true
		   else CrCom.useNonMinimal := false

	  val crResult = 
          case (#useDirect opt3, #useCommutative opt3, #useLayer opt3) of
               (false,false,false)
               => applyDirectCheck directCheck rs (0,[]) 
             | (true,false,false)
               => applyDirectCheck directCheck rs (0,[0]) 
             | (false,true,false)
               => applyDirectCheck directCheck rs (0,[2]) 
             | (false,false,true)
               => applyDirectCheck directCheck rs (0,[1]) 
             | (true,true,false)
               => applyDirectCheck directCheck rs (0,[0,2]) 
             | (true,false,true)
               => applyDirectCheck directCheck rs (0,[0,1]) 
             | (false,true,true)
               => applyDirectCheck directCheck rs (0,[1,2]) 
             | (true,true,true)
               => applyDirectCheck directCheck rs (0,[0,1,2]) 
      in
	  case crResult of 
           CR => (print "\nFinal result: CR\n"; CR)
         | NotCR => (print "\nFinal result: not CR\n"; NotCR)
         | Unknown => (print "\nFinal result: Can't judge\n"; Unknown)
      end
      handle Failure => (print "\nFinal result: Can't judge\n"; Unknown)

    end (* of local *)

end (* of structre *)

